/*
	filename here

	Copyright (c) 2000, Peter Siebold

	This file is part of the aems555 project.

	aems555 software is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	aems555 software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with any aems555 software.  If not, see <http://www.gnu.org/licenses/>.
	
	Many thanks to Peter Siebold for allowing us to redistribute his code under GPL!
*/


/* sci. c
	Contains all routines acting directly with reference to the QSM. 
    Uses qsm.h, see this file for function prototypes
*/

#include "hw.h"
#include "qsm.h"  
        


/* 	adapted from Motorola AN   
	Interrupt driven C language routines to implement circular receive 
	and transmit buffers for SCI channel of QSM. As 100% C code is used, 
	the rx and tx buffers can be easily re-sized, or directly accessed.

	Platform : MC68332 or similar CPU32 + QSM micrcontroller
	Compiler : HiWare CM68K CPU32 C Compiler ver 2.6 
	Author   : Mark Maiolani, Motorola EKB, April 96

	Revisions: 13/11/95   - initial creation from CPLOT.C routines
	04/04/96   - Structs used for rx and tx queues
	05/04/96   - sci interrupt handler added
	10/04/96   - rn-stat function added, initial test of rx functions 
			queue initialization and status functions generalised 
			to access a specified structure
	11/04/96   - sci tdre interrupt handler added basic rx and tx 
			functionality tested
	17/04/96   - Comment fixes only
	16/07/96   - SCI receiver OR handlers added, Comments updated
	02/13/00   - adapted for the efi332 project and gcc, JSC
	11/18/00   - adapted for the efi555 project, Peter Siebold
*/


/* overlay structure onto QSM registers
*/
QSMREG *qsmreg = (QSMREG *) QSMBASE;


char rx_byte (u_char *rcvbyte)
{
	/* Read a received byte from the Rx buffer */
	/* Return 0 if no byte available, or non zero if byte read OK */


	u_long  rxq_full_out;
	u_short rxin, rxfull, rxout;

	/* Read rx buffer variables into locals 
	*/ 
	rxin   = rxbuff.in;
	rxfull = rxbuff.full;
	rxout  = rxbuff.out;

	if ((rxin!=rxout) || rxfull)				/* IF (rx data avaialable) */
	{
		*rcvbyte = rxbuff.q[rxout];				/* read data byte */
		rxout++;								/* update pointer..*/
		if (rxout>(qsize-1)) rxout =    0;		/* check for wraparound */
		rxq_full_out = (u_long)rxout;			/* prepare for coherent write..*/
		*(u_long *)(&rxbuff.full) = rxq_full_out; 	/* of rxqout with rxqfull zero */ 
		return 1;								/* return I for successful read */
	}
	else
	{
		return 0;								/* return 0 as no byte available */
	}
}


void  qinit (queue_struct *qvar)
{
	/* Initialize specified queue */

	qvar->in   = 0x0002;						/* Set in=out, and full=O, ie. buffer empty */
	qvar->full = 0x0000;
	qvar->out  = 0x0002;
}


u_short qstat (queue_struct *qvar)
{
	/* Return the number of valid bytes in a specified buffer */

	u_short qin, qfull, qout;

	/* Read q buffer variables into locals 
	*/ 
	qin   = qvar->in;
	qfull = qvar->full;
	qout  = qvar->out;

	if (qin>qout)
	{
		/* Data in buffer, no pointer wrap
		*/
		return(qin-qout);
	}
	if (qin<qout)
	{
		/* Data in buffer with pointer wrap
		*/
		return(qsize-qout+qin);
	}

	/* qin must be = qout
	*/
	if (qfull)
	{
		/* Buffer full 
		*/ 
		return (qsize);
	}
	/* Buffer empty
	*/
	return (0);
}



char tx_byte (u_char txbyte)
{
	/* Queue a byte into the TX buffer for transmission */
	/* Return 0 if buffer already full, or non zero if byte queued 0X */

	u_short txin, txfull, txout;
	u_long  txq_in_full;

	/* Read rx buffer variables into locals 
	*/ 
	txin   = txbuff.in;
	txfull = txbuff.full;
	txout  = txbuff.out;

	if ((txin!=txout)||!txfull) 				/* IF (tx buffer not full) */

		/* Add data to tx buffer as space is available */
	{
		txbuff.q[txin] = txbyte;
		txin++;
		if (txin>(qsize-1)) txin = 0; 			/* check for wraparound */

		if (txin==txout)
			/* Buffer is now full, need to set FULL flag coherently with update of txin */ 
			/* Also generate warning that buffer is now full */
		{
			txq_in_full = (((u_long)txout)<<16)|1; 		/* prepare for coherent write..*/ 
			*(u_long *)(&txbuff.in) = txq_in_full;		/* of txin with txqfull non-zero */
		}
		else
		{
			/* Buffer not full yet, so don't set FULL 
			*/ 
			txbuff.in = txin;
		}

		/* Enable transmitter interrupt (TIE bit set) 
		*/ 
		qsmreg->scc2r1.TE = 1;    	/* tx enable  */
		qsmreg->scc2r1.RE = 1;		/* rx enable  */
		qsmreg->scc2r1.TIE = 1;		/* tx interrupt enable	*/
		qsmreg->scc2r1.RIE = 1;		/* rx interrupt enable	*/
        
		return (1);
	}
	else
		/* Return error code as no buffer space to queue data */ 
	{
		return (0);
	}
}



void SCIinit()
/* Power-up initialization of SCI */
{

	u_char inttmpString[10];
	puts2("\n Init qsm buffers!");  
	
	qinit(&rxbuff);						/* initialize RX buffer */ 
	qinit(&txbuff);						/* initialize TX buffer */
	 											
	/* IARB = see efi.h, supv access only 
	*/
/*	qsmreg->qsmcr.IARB = QSM_IARB;
	qsmreg->qsmcr.SUPV = 1;*/
    

	/* SCI interrupt level and vector number (SCIVEC) 
	*/

    qsmreg->qsciilr.ILSCI = SCI_IRQL;	/* interrupts priority */
/*	qsmreg->qivr = SCIVEC;				/* scivec vector table index, Removed for 555?	*/

	/*
	Select baud rate using:
	Baud rate = system clock / (32 * SCCRO value) 
	*/


/*	qsmreg->scc2r0.SCBR = 130;		//9600	*/
/*	qsmreg->scc2r0.SCBR = 33;		//38400	*/
	qsmreg->scc2r0.SCBR = 22;		//57600	*/
/*	qsmreg->scc2r0.SCBR = 1;		//1250000	*/

		/* 	Enable receiver interrupt (RiE bit set)
		This enables interrupts generated by RDRF and/or OR flags 
	*/

	qsmreg->scc2r1.TE = 1;    	/* tx enable  */
	qsmreg->scc2r1.RE = 1;		/* rx enable  */
	qsmreg->scc2r1.RIE = 1;		/* rx interrupt enable */
}

	
void sciint()
/* SCI RX interrupt handler */

/* 	Note : The SCI interrupt handler will only clear a single interrupt source each time 
	it is executed, by using an else-if structure. This ensures that the transmitter routine 
	is not called unnecessarily if the SCI transmitter is idle (TDRE set) but with no 
	transmit buffer data pending, and the interrupt source therefore disabled 
*/
{
	u_short status, qin, qfull, qout;
	u_short scidata; 
	u_long  q_long;

	/* Read status reg to determine source of exception 
	*/ 
	status = qsmreg->sc2sr;

	if (status & SCSR_RDRF)				/* if RDRF (rx data register full) flag set */
	{
		scidata = qsmreg->sc2dr;			/* Read received data (+ clear RDRF) */

		if (status & SCSR_OR)			/* if OR (receiver OverRun) flag set */

		/* 	Note : typically RDRF will be set at the same time as OR, in which case
			the RDRF handler will clear OR.  OR may be set on its own if an overrun 
			is generated in between QSM_SCSR being read with RDRF set (normal receive), 
			and RDRF being cleared by reading SCDR. Because of this possibility, we 
			need an independent OR flag test and clear mechanism 
		*/
		{

			/*	
				Any additional receiver overrun handling should be added here and in the
				separate OR handler. 
			*/

		}
		else
	 	/* 	Note : This s/w flow discards any received data which has caused
	 		a receiver overrun, as it may be regarded as a non-recoverable error. 
	 		The slw could be madofied by removing the previous else, statement
	 		so that no data is discarded, although this may result in missing data 
	 		from the receive buffer when overrun occurs. 
	 	*/
	 	{
			/* Read rx buffer variables into locals 
			*/
			qin   = rxbuff.in;
			qfull = rxbuff.full;
			qout  = rxbuff.out;

			if ((qin!=qout)||!qfull)   					/* IF (rx buffer not full) */ 
	 		/* data received, but OK. as space in rx buffer 
	 		*/
	 		{
				rxbuff.q[qin] = (u_char)scidata;
				qin++;
				if (qin>(qsize-1)) qin = 0; 			/* check for wraparound */
				if (qin==qout)
	 			/* Buffer now full, so set FULL flag coherently with update of qin */ 
	 			/* Also generate warning that buffer is now full */
	 			{
					q_long = (((u_long)qout)<<16)|1;		/* prepare for coherent write..*/ 
					*(u_long *)(&rxbuff.in) = q_long;	/* of qin with rxqfull non-zero*/
	 			}
				else
	 			{
					/* Buffer not full yet, so don't set FULL 
					*/ 
					rxbuff.in = qin;
	 			}
	 		}
			else
	 		{
				/* Data received, but buffer full (buffer overrun) so flag error */ 
				/* Any additional handler for buffer overrun should be added here */
	 		}
	 	}
    }
	else if (status & SCSR_OR)				 /* if OR (receiver OverRun) flag set */
	/* 	Note : typically RDRF will be set at the same time as OR, in which case
		the RDRF handler will clear OR.  OR may be set on its own if an overrun 
		is generated in between QSM_SCSR being read with RDRF set (normal receive), 
		and RDRF being cleared by reading SCDR. Because of this possibility, we 
		need an independent OR flag test and clear mechanism 
	*/
	{
		scidata = qsmreg->sc2dr;				/* Read received data to clear OR */
		
		/* 	Any additional receiver overrun handling should be added here and in the 
			RDRF + OR handler */
	}
	else if (status & SCSR_TDRE)				/* if TDRE (tx data register empty) flag set */
	{
		qin   = txbuff.in; 
		qfull = txbuff.full; 
		qout = txbuff.out;
    
		if ((qin!=qout)||qfull)					/* IF (tx data available) */
		{
			scidata = txbuff.q[qout];			/* read data byte */
			qout++;								/* update pointer..*/
			if (qout>(qsize-1)) qout = 0;		/* check for wraparound */
			q_long = (u_long)qout;				/* prepare for coherent write..*/ 
			*(u_long *)(&txbuff.full) = q_long;	/* of txqout with txqfull zero */

			qsmreg->sc2dr = scidata;		   		/* Write tx data to SCI (+ clear TDRE)*/
		}
		else									/* no more tx data pending.	 */
		{
			/* ..so disable transmitter interrupt (clear TIE bit) 
			*/ 
			qsmreg->scc2r1.TIE = 0;
		}
	}  
	else									/* some other interrupt source */
	{
	    SCIinit () ;						/* something is wrong - re-initialize sci */
	}
}


void QSPIinit()
/* initialize the QSPI part of the QSM.  Because the SCI and QSPI share registers
   this code assumes that they are used together.  
*/
{
    int i;

    qsmreg->qsmcr.STOP 	 = 0;           /* Stop enable 0: normal QSM clock ops, 1: disable clock ops 		*/
    qsmreg->qsmcr.FRZ    = 1;	 		/* Freeze control 0: ignore freeze, 1: Halt QSPI on xfer boundary	*/

	qsmreg->qspiilr.ILQSPI  = QSPI_IRQL;   /* QSM interrupt priority     			*/
    
    qsmreg->portqs = 0x78;            	/* qsm pin assignments : PSC3-0 		*/
    qsmreg->pqspar = 0x7F;				/* qsm pin assignmants					*/
    qsmreg->ddrqs  = 0x7E;              /* qsm pin assignments, directions  	*/
    
    /* sprc0
    */
    qsmreg->spcr.MSTR = 1;              /* 0: QSPI slave, 1: QSPI master		 		*/
    qsmreg->spcr.BITS = 10;             /* Bits per xfer: 10 bit xfers  		 		*/
    qsmreg->spcr.CPOL = 0;              /* clock polarity, inactive state value 0,1    	*/
    qsmreg->spcr.CPHA = 0;              /* clock phase: 0: Capture data on leading edge of SCK, change on falling edge 	 */
                                        /*              1: Change data on leading edge, capture on following edge of SCK */

    qsmreg->spcr.SPBR = qspidivider(QSPI_BAUD);
    
    /* spcr1
    */
    qsmreg->spcr.DSCKL = 71;            /* length of delay from PCS valid to SCK transition    */
    qsmreg->spcr.DTL   = 16;            /* length of delay after xfer       */
    
    /* sprc2
    */
    qsmreg->spcr.SPIFIE = 0;   			/* QSPI interrupts 0: disabled, 1: enabled		 */
    qsmreg->spcr.WREN   = 1;   			/* Wrap Enable     0: disabled, 1: enabled		 */
    qsmreg->spcr.WRT0 	= 1;   	        /* Wrap To         0: $0,       1, wrap to NEWQP */
    qsmreg->spcr.ENDQP  = 12;           /* start queue address   			             */
    qsmreg->spcr.NEWQP  = 1;            /* end queue address                			 */

    /* spcr 3 
    */
    qsmreg->spcr.HALT = 0;				/* Halt		0: disabled, 1: enabled			*/
	qsmreg->spcr.SPIF = 0;				/* QSPI finished flag, 0: nay, 1: yay		*/
    
    /* Serial Switch (PCS0)
    */
   	qsmreg->cr[0].CONT  = 0;		/* 0: after xfer, PCS use PORTQS, 1: PCS remain asserted						*/
   	qsmreg->cr[0].BITSE = 0;		/* 0: use 8 bit xfer, 1: use BITS in qsmreg->spcr.BITS							*/
    qsmreg->cr[0].DT    = 0;		/* 0: no delay, 1: use qsmreg->spcr.DTL for delay after xfer					*/
   	qsmreg->cr[0].DSCK  = 0;		/* 0: PCS valid to SCK transition is 1/2 SCK, 1: qsm->spcr.DSCKL specs delay	*/ 
   	qsmreg->cr[0].PCS   = 14;		/* Peripheral Chip Select: PCS3, PCS2, PCS1														*/
   
    /* QSPI Command Ram A/D #1 Configuration
    **		ANO	PCS1	-	MAP, 	goes to qrec[3]					
    **		AN1 PCS1	-	BARO, 	goes to qrec[5]
    **		AN2 PCS1	-	MAT, 	goes to qrec[7]
    **		AN3 PCS1	-	ACCELG,	goes to qrec[9]
    **		AN4 PCS1	-	BAT, 	goes to qrec[11]
    **		AN5 PCS1	-	Spare, 	goes to qrec[1]
    **		
    */
    for(i = 1; i <= 11; i+=2)
    {
    	qsmreg->cr[i].CONT  = 0;		/* 0: after xfer, PCS use PORTQS, 1: PCS remain asserted						*/
        qsmreg->cr[i].BITSE = 1;		/* 0: use 8 bit xfer, 1: use BITS in qsmreg->spcr.BITS							*/
	    qsmreg->cr[i].DT    = 0;		/* 0: no delay, 1: use qsmreg->spcr.DTL for delay after xfer					*/
    	qsmreg->cr[i].DSCK  = 1;		/* 0: PCS valid to SCK transition is 1/2 SCK, 1: qsm->spcr.DSCKL specs delay	*/ 
    	qsmreg->cr[i].PCS   = 13;		/* Peripheral Chip Select: PCS3, PCS2, PCS0										*/
    }

    /* QSPI Command Ram A/D #2 Configuration
    **		ANO	PCS3	-	TPS, 	goes to qrec[4]					
    **		AN1 PCS3	-	COOL, 	goes to qrec[6]
    **		AN2 PCS3	-	KNOCK,	goes to qrec[8]
    **		AN3 PCS3	-	Spare,	goes to qrec[10]
    **		AN4 PCS3	-	O2, 	goes to qrec[12]
    **		AN5 PCS3	-	Spare, 	goes to qrec[2]
    **		
    */
	for(i = 2; i <= 12; i+=2)
    {
    	qsmreg->cr[i].CONT  = 0;		/* 0: after xfer, PCS use PORTQS, 1: PCS remain asserted						*/
	   	qsmreg->cr[i].BITSE = 1;		/* 0: use 8 bit xfer, 1: use BITS in qsmreg->spcr.BITS							*/
	    qsmreg->cr[i].DT    = 1;		/* 0: no delay, 1: use qsmreg->spcr.DTL for delay after xfer					*/
        qsmreg->cr[i].DSCK  = 1;		/* 0: PCS valid to SCK transition is 1/2 SCK, 1: qsm->spcr.DSCKL specs delay	*/ 
    	qsmreg->cr[i].PCS   = 7;		/* Peripheral Chip Select: PCS2, PCS1, PCS0										*/
    }
	
    /* spare command ram init (not necessary)
    */
	for(i = 13; i <= 15; i++)
    {
    	qsmreg->cr[i].CONT  = 0;		/* 0: after xfer, PCS use PORTQS, 1: PCS remain asserted						*/
    	qsmreg->cr[i].BITSE = 0;		/* 0: use 8 bit xfer, 1: use BITS in qsmreg->spcr.BITS							*/
	    qsmreg->cr[i].DT    = 0;		/* 0: no delay, 1: use qsmreg->spcr.DTL for delay after xfer					*/
        qsmreg->cr[i].DSCK  = 0;		/* 0: PCS valid to SCK transition is 1/2 SCK, 1: qsm->spcr.DSCKL specs delay	*/ 
    	qsmreg->cr[i].PCS   = 0;		/* Peripheral Chip Select: PCS2, PCS1, PCS0										*/
	}    
    
    /* QSPI transmit ram
    **      addresses for A/D channels, 4 msb of 10 bit word, 0,0, 40,40, 80,80, ...
    **
    */				
    qsmreg->tr[0]  = 0xFF;
    for(i = 1; i <= 12; i++)
        qsmreg->tr[i] = ((i - 1) / 2) * 64; 
    qsmreg->tr[13] = 0x00;
    qsmreg->tr[14] = 0x00;
    qsmreg->tr[15] = 0x00;    
    
	/* let er rip
    */
    qsmreg->spcr.SPE   = 1;             /* enable QSPI  */
    
	/* wait until QSPI commands sent to all 12 A/D channels
    */
    while(qsmreg->spcr.SPIF == 0);
    
    /* wait until QSPI cmd sent to channel 1 again to ensure all data valid
    */
    while(qsmreg->spcr.CPTQP == 0);
                     
}              
                     
                     

void SCIPutInt(int num, int len)
{
unsigned char *bptr = &locbuf[10];

int negative, count = 0;

	if (len >= 10) return;
	if ((negative = (num < 0))) num = -num;
	if (num == 0) {
		*--bptr = '0';
	} else {
		while (num && (bptr != locbuf)) {
			*--bptr = '0' + (num % 10);
			num /= 10;
			count++;
		}
	}
	if (negative && (bptr != locbuf)) {*--bptr = '-'; count++;}
	while (count < len) {*(--bptr) = ' '; count++;}
	for (negative = 0; negative < count; negative++)
		SCIPutBytes(&bptr[negative], 1);
}

void SCIPutHex(unsigned int num, int len)
{
unsigned char *bptr = &locbuf[10];
int count = 0, idx;
unsigned char c;

	if (len >= 10) return;
	if (num == 0) {
		*--bptr = '0';
	} else {
		while (num && (bptr != locbuf)) {
			c = (num % 16);
			if (c > 9) c += ('A' - 10); else c += '0';
			*--bptr = c;
			num /= 16;
			count++;
		}
	}
	while ((bptr != locbuf) && (count < len)) { *(--bptr) = '0'; count++; }
	SCIPutS("0x");
	for (idx = 0; idx < count; idx++)
		SCIPutBytes(&bptr[idx], 1);
}
                     

void SCIPutS(char *message)
{
	/* while not null load message into the transmit buffer
	*/
	while(*message)
	{
		tx_byte(*message);
		message++;
	}
}


void SCIPutBytes(char *val, int len)
{
	int i;
    
	for( i=0; i < len; i++)
    {
		tx_byte(*val);
        val++;
	}
}


    	

